@@ -1,7 +1,7 @@
 package Params::Sah;
 
-our $DATE = '2015-04-22'; # DATE
-our $VERSION = '0.04'; # VERSION
+our $DATE = '2016-06-02'; # DATE
+our $VERSION = '0.05'; # VERSION
 
 use 5.010001;
 use strict;
@@ -46,7 +46,7 @@ sub gen_validator {
     my $src = '';
 
     my $i = 0;
-    my %mentioned_mods;
+    my @modules_for_all_args;
     my %mentioned_vars;
 
     # currently prototype won't force checking
@@ -78,18 +78,13 @@ sub gen_validator {
             return_type  => $return_type,
             indent_level => 1,
         );
-        for my $mod (sort keys %{ $cd->{module_statements} }) {
-            next if $mentioned_mods{$mod}++;
-            my $stmt = $cd->{module_statements}{$mod};
-            $src .= "    $stmt->[0] $mod".
-                ($stmt->[1] && @{ $stmt->[1] } ?
-                     " ".join(", ", @{ $stmt->[1] }) : "").";\n";
-        }
-        for my $mod (@{ $cd->{modules} }) {
-            next if $cd->{module_statements}{$mod};
-            (my $mod_pm = $mod) =~ s!::!/!g; $mod_pm .= ".pm";
-            next if $mentioned_mods{$mod}++;
-            require $mod_pm;
+        die "Incompatible Data::Sah version (cd v=$cd->{v}, expected 2)" unless $cd->{v} == 2;
+        for my $mod_rec (@{ $cd->{modules} }) {
+            next unless $mod_rec->{phase} eq 'runtime';
+            next if grep { ($mod_rec->{use_statement} && $_->{use_statement} && $_->{use_statement} eq $mod_rec->{use_statement}) ||
+                               $_->{name} eq $mod_rec->{name} } @modules_for_all_args;
+            push @modules_for_all_args, $mod_rec;
+            $src .= "    ".($mod_rec->{use_statement} // "require $mod_rec->{name}").";\n";
         }
         for my $var (sort keys %{$cd->{vars}}) {
             next if $mentioned_vars{$var}++;
@@ -150,7 +145,7 @@ Params::Sah - Validate method/function parameters using Sah schemas
 
 =head1 VERSION
 
-This document describes version 0.04 of Params::Sah (from Perl distribution Params-Sah), released on 2015-04-22.
+This document describes version 0.05 of Params::Sah (from Perl distribution Params-Sah), released on 2016-06-02.
 
 =head1 SYNOPSIS
 
@@ -199,47 +194,6 @@ Performance is faster than Params::Validate, since you can avoid recompiling
 specification or copying array/hash twice. Sah also provides a rich way to
 validate data structures.
 
-=head1 FUNCTIONS
-
-None exported by default, but exportable.
-
-=head2 gen_validator([\%opts, ] ...) => code
-
-Generate code for subroutine validation. It accepts an optional hashref as the
-first argument for options. The rest of the arguments are Sah schemas that
-correspond to the function parameters in the same position, i.e. the first
-schema will validate the function's first argument, and so on. Example:
-
- gen_validator('schema1', 'schema2', ...);
- gen_validator({option=>'val', ...}, 'schema1', 'schema2', ...);
-
-Will return a coderef which is the validator code. The code accepts an arrayref
-(usually C<< \@_ >>).
-
-Known options:
-
-=over
-
-=item * named => bool (default: 0)
-
-If set to true, it means we are generating validator for subroutine that accepts
-named parameters (e.g. C<< f(name=>'val', other=>'val2') >>) instead of
-positional (e.g. C<< f('val', 'val2') >>). The validator will accept the
-parameters as a hashref. And the arguments of C<gen_validator> are assumed to be
-a hash of parameter names and schemas instead of a list of schemas, for example:
-
- gen_validator({named=>1}, arg1=>'schema1', arg2=>'schema2', ...);
-
-=item * on_invalid => str (default: 'croak')
-
-What should the validator code do when function parameters are invalid? The
-default is to croak (see L<Carp>) to report error to STDERR from the caller
-perspective. Other valid choices include: C<warn>, C<carp>, C<die>, C<bool>
-(return false on invalid, or true on valid), C<str> (return an error message on
-invalid, or empty string on valid).
-
-=back
-
 =head1 PERFORMANCE NOTES
 
 Sample benchmark against Params::Validate:
@@ -249,6 +203,8 @@ Sample benchmark against Params::Validate:
  use 5.010;
  use strict;
  use warnings;
+ use FindBin '$Bin';
+ use lib "$Bin/../lib";
  
  use Benchmark::Dumb qw(cmpthese);
  use Params::Sah qw(gen_validator);
@@ -311,6 +267,47 @@ Sample benchmark result on my laptop:
  P::Sah, pos, str+int      837250+-880/s          973.5+-1.1%        670.2+-1.3%  197.33+-0.64%            82.63+-0.2%                   --           -23.9%
  P::Sah, pos, str       1.0997e+06+-24/s              1310.0%        911.6+-1.3%  290.54+-0.73%                 139.9%         31.35+-0.14%               --
 
+=head1 FUNCTIONS
+
+None exported by default, but exportable.
+
+=head2 gen_validator([\%opts, ] ...) => code
+
+Generate code for subroutine validation. It accepts an optional hashref as the
+first argument for options. The rest of the arguments are Sah schemas that
+correspond to the function parameters in the same position, i.e. the first
+schema will validate the function's first argument, and so on. Example:
+
+ gen_validator('schema1', 'schema2', ...);
+ gen_validator({option=>'val', ...}, 'schema1', 'schema2', ...);
+
+Will return a coderef which is the validator code. The code accepts an arrayref
+(usually C<< \@_ >>).
+
+Known options:
+
+=over
+
+=item * named => bool (default: 0)
+
+If set to true, it means we are generating validator for subroutine that accepts
+named parameters (e.g. C<< f(name=>'val', other=>'val2') >>) instead of
+positional (e.g. C<< f('val', 'val2') >>). The validator will accept the
+parameters as a hashref. And the arguments of C<gen_validator> are assumed to be
+a hash of parameter names and schemas instead of a list of schemas, for example:
+
+ gen_validator({named=>1}, arg1=>'schema1', arg2=>'schema2', ...);
+
+=item * on_invalid => str (default: 'croak')
+
+What should the validator code do when function parameters are invalid? The
+default is to croak (see L<Carp>) to report error to STDERR from the caller
+perspective. Other valid choices include: C<warn>, C<carp>, C<die>, C<bool>
+(return false on invalid, or true on valid), C<str> (return an error message on
+invalid, or empty string on valid).
+
+=back
+
 =head1 FAQ
 
 =head2 How do I learn more about Sah (the schema language)?
@@ -377,14 +374,6 @@ Sample output:
     |
   26|};
 
-=head1 SEE ALSO
-
-L<Sah>, L<Data::Sah>
-
-L<Params::Validate>
-
-L<Perinci::Sub::Wrapper>, if you want to do more than parameter validation.
-
 =head1 HOMEPAGE
 
 Please visit the project's homepage at L<https://metacpan.org/release/Params-Sah>.
@@ -401,13 +390,24 @@ When submitting a bug or request, please include a test-file or a
 patch to an existing test-file that illustrates the bug or desired
 feature.
 
+=head1 SEE ALSO
+
+L<Sah>, L<Data::Sah>
+
+L<Params::Validate>
+
+L<Perinci::Sub::Wrapper>, if you want to do more than parameter validation.
+
+L<Perinci::Sub::ValidateArgs> and L<Data::Sah::Params>, an accidental
+reimplementation of the same idea a year later :-)
+
 =head1 AUTHOR
 
 perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by perlancar@cpan.org.
+This software is copyright (c) 2016 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
